import unittest
import pkg_resources
import webtest

from pyramid.config import Configurator
from pyramid.decorator import reify


class Test(unittest.TestCase):

    @reify
    def config(self):
        from example import includeme
        _config = Configurator(settings={})
        _config.include(includeme)
        self.addCleanup(delattr, self, 'config')
        return _config

    @reify
    def app(self):
        self.addCleanup(delattr, self, 'app')
        return webtest.TestApp(self.config.make_wsgi_app())

    def test_root(self):
        self.app.get('/')

    def test_users_index(self):
        self.app.get('/users')

    def test_post_photo(self):
        image_gif = pkg_resources.resource_stream('royal.tests.functional',
                                                  'image.gif')
        result = self.app.post(
            '/users/hadrien/photos',
            upload_files=[(u'image', u'image.gif', image_gif.read())]
            )



from zope.interface import implementer

from pyramid.decorator import reify
from pyramid.traversal import find_root

from royal import exceptions as exc
from royal.interfaces import (
    IBase,
    ICollection,
    IPaginatedResult,
    IResource,
    IRoot,
    )


@implementer(IBase)
class Base(object):

    children = {}

    def __init__(self, name, parent):
        self.__name__ = name
        self.__parent__ = parent

    @reify
    def root(self):
        return find_root(self)

    @property
    def links(self):
        return {'self': self}

    def __getitem__(self, key):
        return self.children[key](key, self)


@implementer(IRoot)
class Root(Base):
    children = {}

    def __init__(self, request):
        super(Root, self).__init__('', None)
        self.request = request

    def show(self):
        return None

    @property
    def links(self):
        return dict((name, cls(name, self))
                    for name, cls in self.children.iteritems())


@implementer(IResource)
class Resource(Base):

    def show(self):
        raise exc.MethodNotAllowed(self)

    def put(self):
        raise exc.MethodNotAllowed(self)

    def patch(self):
        raise exc.MethodNotAllowed(self)

    def delete(self):
        raise exc.MethodNotAllowed(self)

    @reify
    def resource_name(self):
        """Resource name used at rendering phase.
        By default, it is collection name without last s.
        """
        if self.__parent__ is not None:
            return self.__parent__.__name__[0:-1]
        return self.__name__

    @property
    def links(self):
        return {'self': self}


@implementer(ICollection)
class Collection(Base):

    def index(self, *args, **kwargs):
        raise exc.MethodNotAllowed(self)

    def create(self, params):
        raise exc.MethodNotAllowed(self)

    def delete(self):
        raise exc.MethodNotAllowed(self)


@implementer(IPaginatedResult)
class PaginatedResult(object):

    def __init__(self, parent, iterator, resource_cls, query, total):
        self.parent = parent
        self.iterator = iterator
        self.resource_cls = resource_cls
        self.query = query
        self.total = total

    def __iter__(self):
        for item in self.iterator:
            yield self.resource_cls(item._id, self.parent, model=item)



import royal

from pyramid.decorator import reify

from example.resources import user
from example.resources import photo


class Root(royal.Root):

    children = {
        'users': user.Collection,
        'photos': photo.Collection,
        }

    def __getitem__(self, key):
        return self.children[key](key, self)

    @reify
    def db(self):
        return self.request.mongo_connection.get_db()



import os
import logging

from pyramid_mongokit import SingleDbConnection

from example.model import User

log = logging.getLogger(__name__)


def setup_package():
    os.environ['MONGO_URI'] = 'mongodb://localhost'
    os.environ['MONGO_DB_NAME'] = 'royal_example'

    connection = new_connection()
    setup_example_fixtures(connection)
    connection.close()


def teardown_package():
    connection = new_connection()
    for name in connection.db.collection_names():
        if u'system.indexes' not in name:
            connection.db[name].remove()
    connection.close()


def new_connection():
    db_prefix = ''
    return SingleDbConnection(os.environ['MONGO_DB_NAME'], db_prefix,
                              os.environ['MONGO_URI'])


def setup_example_fixtures(connection):
    connection.register(User)
    connection.generate_index(User)
    try:
        User.create(connection.db, u'hadrien', u'hadrien@ectobal.com')
    except Exception:
        log.exception('setup fixture failed')



from royal.resource import Root, Collection, Resource, PaginatedResult


__all__ = ['Root', 'Collection', 'Resource', 'PaginatedResult']


def includeme(config):
    config.include('royal.renderer')
    config.include('royal.views')
    config.commit()



import logging
from functools import wraps

from onctuous import All, Coerce, InRange, Required, Schema

from pyramid.view import view_defaults, view_config
from pyramid.httpexceptions import (
    HTTPCreated,
    HTTPConflict,
    HTTPBadRequest,
    HTTPInternalServerError,
    HTTPNotFound,
    HTTPMethodNotAllowed,
    )

from royal import exceptions as exc
from royal import interfaces

log = logging.getLogger(__name__)


def includeme(config):
    config.scan('royal.views')


class BaseView(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def wrap_dict(self, item, item_dict):
        result = {'links': dict((k, self.request.resource_url(v))
                                for k, v in item.links.iteritems())}
        if item_dict:
            result.update({item.resource_name: item_dict})
        return result

    def resource_url(self, *elements, **kw):
        return self.request.resource_url(self.context, *elements, **kw)


def get_params(request):
    try:
        return request.POST.mixed()
    except Exception:
        log.debug('Error parsing POST', exc_info=True)
        raise HTTPBadRequest('Error parsing request parameters')


@view_defaults(context=interfaces.ICollection, renderer='royal')
class CollectionView(BaseView):

    index_schema = {
        Required('offset', 0): All(Coerce(int), InRange(min=0)),
        Required('limit', 20): All(Coerce(int), InRange(min=1)),
        }

    def get_paginated_links(self, query, total_items):
        links = {}
        links['self'] = self.resource_url(query=query)
        offset = query['offset']
        limit = query['limit']

        query['offset'] = 0
        links['first'] = self.resource_url(query=query)

        offset_last_page = (total_items / limit - 1) * limit

        query['offset'] = max(0, offset_last_page)
        links['last'] = self.resource_url(query=query)

        has_previous = offset > 0
        if has_previous:
            query['offset'] = offset - 1
            links['previous'] = self.resource_url(query=query)

        has_next = total_items > offset + limit
        if has_next:
            query['offset'] = offset + 1
            links['next'] = self.resource_url(query=query)

        return links

    @view_config(request_method='GET', permission='index')
    def index(self):
        schema = Schema({})
        schema.schema.update(CollectionView.index_schema)
        if hasattr(self.context, 'index_schema'):
            schema.schema.update(self.context.index_schema.schema)

        query = schema(dict(self.request.GET.mixed()))
        result = self.context.index(**query)

        items = [self.wrap_dict(item, item.show()) for item in result]

        return {
            self.context.__name__: items,
            'links': self.get_paginated_links(result.query, result.total),
            }

    @view_config(request_method='POST', permission='create')
    def create(self):
        params = get_params(self.request)
        if hasattr(self.context, 'create_schema'):
            params = self.context.create_schema(params)

        item = self.context.create(**params)
        item_url = self.request.resource_url(item)
        self.request.response.headers['Location'] = item_url
        self.request.response.status_int = HTTPCreated.code
        return self.wrap_dict(item, item.show())


@view_defaults(context=interfaces.IResource, renderer='royal')
class ResourceView(BaseView):

    @view_config(request_method='GET', permission='show')
    @view_config(request_method='GET', permission='show',
                 context=interfaces.IRoot)
    def show(self):
        item = self.context.show()
        return self.wrap_dict(self.context, item)

    @view_config(request_method='PUT', permission='put')
    def put(self):
        params = get_params(self.request)
        item = self.context.put(**params)
        return self.wrap_dict(self.context, item)

    @view_config(request_method='PATCH', permission='patch')
    def patch(self):
        self.context.patch(get_params(self.request))
        return self.request.response

    @view_config(request_method='DELETE', permission='delete')
    def delete(self):
        self.context.delete()
        return self.request.response


@view_config(context=exc.MethodNotAllowed)
@view_config(context=interfaces.IBase)
def not_allowed(context, request):
    return HTTPMethodNotAllowed()


def log_error_dict(view_callable):
    @wraps(view_callable)
    def wrapper(context, request):
        result = view_callable(context, request)
        log.debug('%s: %s', type(context), result, exc_info=True)
        return result
    return wrapper


@view_config(context=exc.NotFound, renderer='royal', decorator=log_error_dict)
def item_not_found(context, request):
    request.response.status_int = HTTPNotFound.code
    return {
        'error': 'not_found',
        'resource': request.resource_url(context.resource)
        }


@view_config(context=exc.Conflict, renderer='royal', decorator=log_error_dict)
def conflict(context, request):
    request.response.status_int = HTTPConflict.code
    return {
        'error': 'already_exists',
        'resource': request.resource_url(context.resource)
        }


@view_config(context='onctuous.errors.Invalid',
             renderer='royal', decorator=log_error_dict)
def invalid_parameters(context, request):
    request.response.status_int = HTTPBadRequest.code
    return {
        'error': 'invalid_parameters',
        'message': unicode(context)
        }


@view_config(context=exc.BadParameter, renderer='royal',
             decorator=log_error_dict)
def bad_parameter(context, request):
    request.response.status_int = HTTPBadRequest.code
    return {
        'error': 'invalid_parameters',
        'message': '%s="%s"' % (context.name, context.value)
        }


@view_config(context=Exception, renderer='royal', decorator=log_error_dict)
def exception(context, request):
    request.response.status_int = HTTPInternalServerError.code
    return {
        'error': 'unexpected_error',
        'message': unicode(context),
        'error_class': type(context).__name__,
        }



from pyramid.config import Configurator


def main(global_config, **settings):  # pragma no cover
    config = Configurator(settings=settings)
    config.include(includeme)
    return config.make_wsgi_app()


def includeme(config):
    config.include('royal')
    config.include('example.model')
    config.include('example.resources')





class ExceptionBase(Exception):

    @property
    def resource(self):
        return self.args[0]


class Conflict(ExceptionBase):
    pass


class NotFound(ExceptionBase):
    pass


class MethodNotAllowed(ExceptionBase):
    pass


class BadParameter(ExceptionBase):

    @property
    def name(self):
        return self.args[1]

    @property
    def value(self):
        return self.args[2]



import mimetypes

import royal
from royal.exceptions import NotFound

from example.model import Photo


class Collection(royal.Collection):

    def index(self, offset, limit):
        cursor = Photo.get_newests(self.root.db, offset, limit,
                                   author=self.__parent__.__name__)
        query = dict(offset=offset, limit=limit)
        return royal.PaginatedResult(self, cursor, Resource, query,
                                     cursor.count())

    def create(self, **kwargs):
        fs = kwargs['image']
        author = self.__parent__.__name__
        mime_type = mimetypes.guess_extension(fs.filename)
        photo = Photo.create(self.root.db, author, fs.file, mime_type)
        return Resource(str(photo._id), self, photo)


class Resource(royal.Resource):

    def __init__(self, key, parent, model=None):
        super(Resource, self).__init__(key, parent)
        self.model = model

    def load_model(self):
        if self.model is None:
            self.model = self.root.db.get_by(_id=self.__name__).first()
        if self.model is None:
            raise NotFound(self)
        return self.model

    def show(self):
        return self.load_model()

    @property
    def links(self):
        return {
            'self': self.root['photos'][self.__name__],
            'author': self.__parent__,
        }



# -*- coding: utf-8 -*-
from zope.interface import (
    Attribute,
    Interface,
    )


class IBase(Interface):
    __name__ = Attribute('The __name__ attribute should be the name with which'
                         'a resource’s parent refers to the resource via'
                         '__getitem__.')
    __parent__ = Attribute('Should be a reference to the resource’s parent'
                           'resource instance in the tree')


class IRoot(IBase):
    request = Attribute('The request object.')


class IResource(IBase):

    resource_name = Attribute('Resource name')
    links = Attribute('A mapping which gives relationship with other '
                      'resources.')

    def show():
        "GET /items/{id}"

    def put():
        "PUT /items/{id}"

    def patch():
        "PATCH /items/{id}"

    def delete():
        "DELETE /items/{id}"


class ICollection(IBase):

    def index(offset, limit, *args, **kwargs):
        "GET /items?offset=0&limit=20"

    def create(params):
        "POST /items"

    def delete():
        "DELETE /items"


class IPaginatedResult(Interface):
    total = Attribute('total of items or None')

    def __iter__(self):
        ""
